import abc
import re
from typing import Optional


class ParseError(ValueError):
    pass


class CellParser(abc.ABC):
    """
    A class that can take a string and parse out the CSV cells in the string. String should begin with the first CSV
    text. It can have an arbitrary number of cells afterwards.

    Classes must implement the `parse` method in order to be used for this. See below for a description of the
    signature and contract of that method

    Objects of this class can be passed to the __init__ of CSVParser in order to parse cells of a CSV
    """

    def __init_subclass__(cls) -> None:
        if 'parse' in cls.__dict__:
            raise AttributeError("Classes that subclass 'CellParser' must not override the 'parse' method")
        return super().__init_subclass__()

    @abc.abstractmethod
    def parser_hook(self, text: str) -> tuple[str, int, bool]:
        """
        A method that takes a string of 1 or more CSV cells and parses out the first cell in that string. It returns
        a tuple of the following: 0th - the contents of the first cell in the string `text`; 1st - the first index
        after the end of that cell 2nd - whether or not this the row has ended--i.e. the cell being returned is the
        last cell in a row.

        You must implement this method with the described signature in order to be used in CSVParser
        :param text: the string containing at least 1 CSV cell which must begin at index 0 of the string
        :return: a tuple containing the cell content, first index in the string after the cell, and whether
                 the cell being returned is the final cell in a row
        """
        pass

    def parse(self, text: str) -> tuple[str, int, bool]:
        """
        Called in the CSVParser so that the superclass CellParser can intervene before returning the values to
        the CSVParser. Classes should *NOT* override this method and instead should override parser_hook.
        """
        text, end, done = self.parser_hook(text)
        return text, end, done


class PlainCellParser(CellParser):
    """
    Parses cells that contain no special characters and are not quoted.
    """

    def parser_hook(self, text: str) -> tuple[str, int, bool]:
        def index(string: str, pattern: str, start: Optional[int] = 0):
            try:
                return string.index(pattern, start)
            except ValueError:
                return float('inf')

        comma = index(text, ',')
        newline = index(text, '\n')
        eot = index(text, '\0')
        end = min(comma, newline, eot)
        if end == float('inf'):
            end = len(text)
        content = text[0:end]
        row_done = not (comma < eot and comma < newline)
        end_index = end + 1
        return content, end_index, row_done


class QuotedCellParser(CellParser):
    """
    Parses cells that are quoted by double quotes and can contain special characters including new line but NOT including
    additional double quotes. Empty quoted cells are also allowed
    """

    def parser_hook(self, text: str) -> tuple[str, int, bool]:
        quoted_cell = re.compile(r'"(.*?)"([,\n\x00])', re.DOTALL | re.MULTILINE)
        value = quoted_cell.match(text)
        if value is None:
            raise ParseError("Cannot match quoted text in string starting at {}".format(0))

        content = value.group(1)
        end_index = value.span()[1]
        row_done = value.group(2) == '\n' or value.group(2) == '\x00'
        return content, end_index, row_done


class DefaultCellParser(QuotedCellParser, PlainCellParser):
    """
    Default parser for the CSVParser class: can parse either quoted or plain cells depending on the first character of
    the text that is passed. 
    """

    def parser_hook(self, text: str) -> tuple[str, int, bool]:
        if text[0] == '"':
            return QuotedCellParser.parser_hook(self, text)
        else:
            return PlainCellParser.parser_hook(self, text)



import abc
from typing import Any


class CellFormatter(abc.ABC):
    """
    Subclasses of this class can be used with CSVParser to determine the output of the parser on cells in the
    CSV file. A CellFormatter object passed to the constructor of a CSVParser will be given a chance to format
    the resulting text data of each text when it is parsed by the parser. After parsing out a text, the format
    method of the parser's CellFormatter object is called and then the content is placed in the data structure that
    is returned by the CSVParser object. """

    @abc.abstractmethod
    def format(self, content: str) -> Any:
        """Receives a single text's content during parsing of a CSV in CSVParser and should return the
        the text's content post formatting. Formatting can be arbitrary.

        A class should implement this method in order to be passed to the __init__ of CSVParser. Inheriting
        this class is not required, but it is recommended

        :param content: the content of the CSV text parsed by the CSVParser
        :raises ParseError: if the method cannot parse the content string. If CSVParser sets self.debug to True,
                    then this error is caught and a new error with more information is raised `from` this error.
                    Otherwise, this error is raised unchanged
        :returns: the result of formatting the text
        """
        pass


class DefaultCellFormatter(CellFormatter):
    """
    Does not change the contents of a CSV text. Exists as the default option in CSVParser
    """

    def format(self, content: str) -> str:
        return content


class CleanedCellFormatter(CellFormatter):
    """
    Removes whitespace other than a plain space and double quotes from the content of a CSV text. Since newlines
    are supported inside of double quoted cells by the CSVParser, this will remove them among other the other things
    listed.
    """

    def format(self, content: str) -> str:
        def cleaned(s: str) -> str:
            return s.replace('\r', '').replace('\n', '').replace('\t', '').replace('"', '')

        return cleaned(content)



import abc
from typing import TypeVar


class OutputAdapter(abc.ABC):
    """
    Output adapters take the output of CSVParser and adapt it to a particular format. They are NOT passed into the
    CSVParser but can be instantiated and used outside of the class. They should implement a `adapt` method that
    takes a single list[list[str]] as is returned from the `parse` method of CSVParser. It can return any type.
    """

    def __init__(self):
        pass

    T = TypeVar('T')

    @abc.abstractmethod
    def adapt(self, output: list[list[str]]) -> T:
        """
        takes a single list[list[str]] as is returned from the `parse` method of CSVParser. It can return any type.
        :param output: the output from the CSVParser to be adapted
        :return: The transformed data
        """
        pass


class DictOutputAdapter(OutputAdapter):
    """
    Transforms the CSV output into a list of dictionaries. Each dictionary in the list has the same keys which correspond
    to the values in the first row of the CSV (the header row) and whose keys are one of the entries in the CSV file
    """
    def adapt(self, output: list[list[str]]) -> list[dict[str, str]]:
        headers = output[0]
        entries = output[1:]
        result = []
        for entry in entries:
            result.append({k: v for k, v in zip(headers, entry)})
        return result



from typing import TextIO

from csvtea.cells import CellParser, DefaultCellParser, ParseError
from csvtea.format import DefaultCellFormatter, CellFormatter


class CSVParser:
    def __init__(
            self,
            file: TextIO,
            parser: CellParser = DefaultCellParser(),
            formatter: CellFormatter = DefaultCellFormatter(),
            *,
            debug: bool = False
    ):
        self.parser = parser
        self.formatter = formatter
        self.debug = debug

        self.text = file.read().strip('\r\n') + '\0'

    def parse(self) -> list[list[str]]:
        cells = []
        row = []
        text = self.text
        while text:
            try:
                cell, cell_end, row_done = self.parser.parse(text)
            except ParseError as e:
                if self.debug:
                    error = ParseError(f'Error in parse. Cannot match quoted key. '
                                       f'Text is {text!r}\ncells length is {len(cells)}\n'
                                       f'row length is {len(row)}')
                    raise error from e
                else:
                    raise
            cell = self.formatter.format(cell)
            row.append(cell)
            if row_done:
                cells.append(row)
                row = []
            text = text[cell_end:]
        return cells
